/****************************************************************************
 *
 * geo1: Polymorphe geometrische Klassen
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 4.6: Polymorphie
 *  
 *  - BEACHTE: Unter den Copyrights gibt es Compiler-Switches
 *
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0


// fuer jeden Compiler das seine:
//#define COMPILER_KANN_EXCEPTION_HANDLING
//#define COMPILER_ZU_BLOED_FUER_DEFAULT_ARGUMENTE


/****************************************************************************
 *
 *	Koord.h
 *
 ****************************************************************************/

#ifndef _Koord_h
#define _Koord_h

// Header-Datei fuer I/O
#include <iostream.h>

/* Klasse Koord
 *  - Hilfsklasse fuer geometrische Objekte
 *  - nicht zur Vererbung geeignet
 */
class Koord {
  private:
    int x;              // X-Koordinate
    int y;              // Y-Koordinate

  public:
    // Default-Konstruktor und Konstruktor aus zwei int's
    Koord () {
        x = y = 0;      // Default-Werte: 0
    }
    Koord (int newx, int newy) {
        x = newx;
        y = newy;
    }

    Koord operator+ (const Koord&) const;        // Addition
    Koord operator- () const;                    // Negation
    void operator += (const Koord&);             // +=
    void printOn (ostream& strm = cout) const;   // Ausgabe
};

/* operator+
 *  - X- und Y-Koordinaten addieren
 */
inline Koord Koord::operator+ (const Koord& p) const
{
    return Koord (x+p.x, y+p.y);
}

/* einstelliger operator-
 *  - X- und Y-Koordinaten negieren
 */
inline Koord Koord::operator- () const
{
    return Koord (-x, -y);
}

/* operator+=
 *  - Offset auf X- und Y-Koordinaten aufaddieren
 */
inline void Koord::operator+= (const Koord& p)
{
    x += p.x;
    y += p.y;
}

/* printOn
 *  - Koordinaten als Wertepaar ausgeben
 */ 
inline void Koord::printOn (ostream& strm) const
{
    strm << '(' << x << ',' << y << ')';
}

/* operator<<
 *  - Umsetzung fuer Standard-Ausgabeoperator
 */
inline ostream& operator<< (ostream& strm, const Koord& p)
{
    p.printOn (strm);
    return strm;
}

#endif


/****************************************************************************
 *
 *	GeoObj.h
 *
 ****************************************************************************/

#ifndef _GeoObj_h
#define _GeoObj_h

// Header-Datei fuer Koord
//#include "Koord.h"

/* abstrakte Basisklasse GeoObj
 *  - gemeinsame Basisklasse fuer geometrische Objekte
 *  - zur Vererbung vorgesehen
 */
class GeoObj {
  protected:
    // Jedes GeoObj hat einen Referenzpunkt
    Koord refpunkt;

    /* Konstruktor fuer Startwert vom Referenz-Punkt
     *  - nichtoeffentlich
     *  - damit ist kein Default-Konstruktor vorhanden
     */
    GeoObj (const Koord& p) : refpunkt(p) {
    }

  public:
    // GeoObj um relativen Offset verschieben
    virtual void move (const Koord& offset) {
        refpunkt += offset;
    }
    /* GeoObj ausgeben
     *  - rein virtuelle Funktion
     */
    virtual void draw () const = 0;

    // virtueller Destruktor
    virtual ~GeoObj () {
    }
};

#endif  /* _GeoObj_h */


/****************************************************************************
 *
 *	Kreis.h
 *
 ****************************************************************************/

#ifndef _Kreis_h
#define _Kreis_h

// Header-Datei der Basisklasse
//#include "GeoObj.h"

// Header-Datei fuer I/O
#include <iostream.h>

/* Klasse Kreis
 *  - abgeleitet von GeoObj
 *  - ein Kreis besteht aus:
 *       - Mittelpunkt (Referenzpunkt, geerbt)
 *       - Radius (neu)
 */
class Kreis : public GeoObj {
  protected:
    unsigned radius;                // Radius

  public:
    // Konstruktor fuer Mittelpunkt und Radius
    Kreis (const Koord& m, unsigned r)
         : GeoObj(m), radius(r) {
    }

    // Ausgabe (jetzt auch implementiert)
    virtual void draw () const;

    // virtueller Destruktor
    virtual ~Kreis () {
    }
};

/* Ausgabe
 *  - inline definiert
 */
inline void Kreis::draw () const
{
    cout << "Kreis um Mittelpunkt " << refpunkt
         << " mit Radius " << radius << endl;
}

#endif


/****************************************************************************
 *
 *	Linie.h
 *
 ****************************************************************************/

#ifndef _Linie_h
#define _Linie_h

// Header-Datei der Basisklasse
//#include "GeoObj.h"

// Header-Datei fuer I/O
#include <iostream.h>

/* Klasse Linie
 *  - abgeleitet von GeoObj
 *  - ein Linie besteht aus:
 *       - einem Anfangspunkt (Referenzpunkt, geerbt)
 *       - einem Endpunkt (neu)
 */
class Linie : public GeoObj {
  protected:
    Koord p2;                  // zweiter Punkt, Endpunkt

  public:
    // Konstruktor fuer Anfangs- und Endpunkt
    Linie (const Koord& a, const Koord& b)
         : GeoObj(a), p2(b) {
    }

    // Ausgabe (jetzt auch implementiert)
    virtual void draw () const;

    // Verschieben (neu implementiert)
    virtual void move (const Koord&);

    // virtueller Destruktor
    virtual ~Linie () {
    }
};

/* Ausgabe
 *  - inline definiert
 */
inline void Linie::draw () const
{
    cout << "Linie von " << refpunkt << " bis " << p2 << endl;
}

/* Verschieben
 *  - inline neu implementiert
 */
inline void Linie::move (const Koord& offset)
{
    refpunkt += offset;        // entspricht "GeoObj::move(offset);"
    p2 += offset;
}

#endif


/****************************************************************************
 *
 *	GeoGruppe.h
 *
 ****************************************************************************/

#ifndef _GeoGruppe_h
#define _GeoGruppe_h

#ifdef COMPILER_KANN_EXCEPTION_HANDLING
#else
#include <stdlib.h>
void SizeExceeded() {
    cerr << "Exception: SizeExceeded()" << endl;
    exit(1);
}
#define throw
#endif

#ifdef COMPILER_ZU_BLOED_FUER_DEFAULT_ARGUMENTE
const Koord ursprung (0,0);
#endif

// Header-Datei der Basisklasse einbinden
//#include "GeoObj.h"

/* Klasse GeoGruppe
 *  - abgeleitet von GeoObj
 *  - eine GeoGruppe besteht aus:
 *       - einem Referenzpunkt (geerbt)
 *       - einer Menge von geometrischen Elementen (neu)
 */
class GeoGruppe : public GeoObj {
  protected:
    enum { MAXELEMS = 100 };     // Maximum moeglicher Elemente
    GeoObj*  elems[MAXELEMS];    // Menge von Zeigern auf GeoObj's
    unsigned num;                // Anzahl der Elemente

  public:
    // Konstruktor mit optionalem Referenzpunkt
#ifdef COMPILER_ZU_BLOED_FUER_DEFAULT_ARGUMENTE
    GeoGruppe (const Koord& p = ursprung) : GeoObj(p) {
#else
    GeoGruppe (const Koord& p = Koord(0,0)) : GeoObj(p) {
#endif
        num = 0;
    }

    // Ausgabe (jetzt auch implementiert)
    virtual void draw () const;

    // Element einfuegen
    virtual void add (GeoObj&);

    // virtueller Destruktor
    virtual ~GeoGruppe () {
    }

    // Fehlerklasse
    class SizeExceeded {
    };
};

#endif  /* _GeoGruppe_h */


/****************************************************************************
 *
 *	GeoGruppe.cc / GeoGruppe.cpp
 *
 ****************************************************************************/

/* add
 *  - Element einfuegen
 */
void GeoGruppe::add (GeoObj& obj)
{
    if (num == MAXELEMS) {
        // FEHLER: Groesse ueberschritten
        throw SizeExceeded();
    }

    // Adresse vom geom. Objekt eintragen
    elems[num++] = &obj;
}

/* draw
 *  - alle Elemente unter Beruecksichtigung vom Referenzpunkt ausgeben
 */
void GeoGruppe::draw () const
{
    for (unsigned i = 0; i < num; i++) {
        elems[i]->move (refpunkt);   // Offset fuer Referenzpunkt addieren
        elems[i]->draw ();           // Element ausgeben
        elems[i]->move (-refpunkt);  // Offset wieder entfernen
    }
}


/****************************************************************************
 *
 *	gtest1.cc / gtest1.cpp
 *
 ****************************************************************************/

// Header-Datei fuer I/O
#include <iostream.h>

// Header-Dateien fuer verwendete Klassen
//#include "Linie.h"
//#include "Kreis.h"
//#include "GeoGruppe.h"

main()
{
    Linie l1 (Koord(1,2), Koord(3,4));
    Linie l2 (Koord(7,7), Koord(0,0));
    Kreis k (Koord(3,3), 11);

    GeoGruppe g;

    g.add(l1);            // GeoGruppe enthaelt: - Linie l1
    g.add(k);             //                     - Kreis k
    g.add(l2);            //                     - Linie l2

    g.draw();             // GeoGruppe ausgeben

    g.move(Koord(3,-3));  // GeoGruppen-Offset verschieben
    g.draw();             // GeoGruppe nochmal ausgeben
}

